/********************************************************************************
 *                                  DISCLAIMER                                  *
 * This software and documentation (the "Software") were developed at the Food  *
 * and Drug Administration (FDA) by employees of the Federal Government in the  *
 * course of their official duties. Pursuant to Title 17, Section 105 of the    *
 * United States Code, this work is not subject to copyright protection and is  *
 * in the public domain. Permission is hereby granted, free of charge, to any   *
 * person obtaining a copy of the Software, to deal in the Software without     *
 * restriction, including without limitation the rights to use, copy, modify,   *
 * merge, publish, distribute, sublicense, or sell copies of the Software or    *
 * derivatives, and to permit persons to whom the Software is furnished to do   *
 * so. FDA assumes no responsibility whatsoever for use by other parties of     *
 * the Software, its source code, documentation or compiled executables, and    *
 * makes no guarantees, expressed or implied, about its quality, reliability,   *
 *  or any other characteristic. Further, use of this code in no way implies    *
 * endorsement by the FDA or confers any advantage in regulatory decisions.     *
 * Although this software can be redistributed and/or modified freely, we       *
 * ask that any derivative works bear some notice that they are derived from    *
 * it, and any modified versions bear some notice that they have been modified. * 
 *                                                                              *
 * Written by:                                                                  *
 *  Lars Johannesen                                                             *
********************************************************************************/
#include "mex.h"

#include <stdexcept>
#include <cstdlib>

#include <armadillo>
using namespace arma;

template<typename TYP>
mxArray *set_matlab_scalar(TYP t, mxClassID cla=mxDOUBLE_CLASS) {
        mwSize dims[2] = {1,1};
        mxArray *o = mxCreateNumericArray(2, dims, cla, mxREAL);
        double *d = (double*)mxGetData(o);
        *d=t;
        return o;
}

void mexFunction (int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	if(nrhs != 3)
		mexErrMsgTxt("Need 3 arguments: refannmat, annmat, threshold");

	try {
		if(!mxIsDouble(prhs[0]))
			mexErrMsgTxt("First argument was not a double");

		if(!mxIsDouble(prhs[1]))
			mexErrMsgTxt("Second argument was not a double");

		if(!mxIsDouble(prhs[2]))
			mexErrMsgTxt("Third argument was not a double scalar");

		double *ptr = (double*)mxGetPr(prhs[0]);
		int cols = mxGetN(prhs[0]);
		int rows = mxGetM(prhs[0]);
		if(cols != 1) {
			mexErrMsgTxt("First argument must be nx1");
		}
		mat refann(ptr, rows, cols);

		ptr = (double*)mxGetPr(prhs[1]);
		cols = mxGetN(prhs[1]);
		rows = mxGetM(prhs[1]);
		if(cols != 1) {
			mexErrMsgTxt("Second argument must be nx1");
		}
		mat ann(ptr, rows, cols);

		ptr = (double*)mxGetPr(prhs[2]);
		double threshold = *ptr;

		// Store annots
		mxArray *mmat = mxCreateDoubleMatrix(refann.n_rows,1,mxREAL);
		double *d = (double*) mxGetData(mmat);
		mat omat(d, refann.n_rows, 1, false, true);

		int tp=0;
		int fp=0;
		int fn=0;

		const int HIT=-1;

		for(std::size_t i=0; i < refann.n_rows; ++i) {
			bool found = false;
			std::size_t bestj = 0;
			double bestdiff = 0;

			if(mxIsNaN(refann(i,0))) continue;

			for(std::size_t j=0; j < ann.n_rows; ++j) {
				if(!mxIsNaN(ann(j,0)) && ann(j,0) != HIT) {
					// Unmatched entry

					if(!found) {
						double diff = fabs(ann(j,0)-refann(i,0));

						if(diff < threshold) {
							found = true;
							bestdiff = diff;
							bestj = j;
						}
					} else {
						double diff = fabs(ann(j,0)-refann(i,0));

						if(diff < bestdiff) {
							bestdiff = diff;
							bestj = j;
						}
					}
				}
			}

			if(found) {
				++tp; // We got it
				omat(i,0) = bestdiff;
				ann(bestj,0) = HIT;
				refann(i,0) = HIT;
			} else {
				++fn;
				omat(i,0) = mxGetNaN();
			}
		}

		for(std::size_t j=0; j < ann.n_rows; ++j) {
			if(!mxIsNaN(ann(j,0)) && ann(j,0) != HIT) {
				++fp;
			}
		}

		plhs[0] = set_matlab_scalar(tp);

		if(nlhs > 0) {
			plhs[1] = set_matlab_scalar(fp);
		}

		if(nlhs > 1) {
			plhs[2] = set_matlab_scalar(fn);
		}

		if(nlhs > 2) {
			plhs[3] = mmat;
		}

		if(nlhs > 3) {
			mwSize dims[2];
			dims[0] = refann.n_rows;
			dims[1] = refann.n_cols;
			mxArray *o = mxCreateNumericArray(2, dims, mxDOUBLE_CLASS, mxREAL);
			ptr = (double*)mxGetData(o);

			memcpy(ptr, refann.memptr(), refann.n_rows * refann.n_cols * sizeof(double));

			plhs[4] = o;
		}

		if(nlhs > 4) {
			mwSize dims[2];
			dims[0] = ann.n_rows;
			dims[1] = ann.n_cols;
			mxArray *o = mxCreateNumericArray(2, dims, mxDOUBLE_CLASS, mxREAL);
			ptr = (double*)mxGetData(o);

			memcpy(ptr, ann.memptr(), ann.n_rows * ann.n_cols * sizeof(double));

			plhs[5] = o;
		}
	} catch (std::logic_error &exp) {
		mexErrMsgTxt(exp.what());
	} catch (...) {
		mexErrMsgTxt("Unknown error");
	}
}
