/**
 * @file annotation.hpp
 * @author Lars Johannesen <lj@halloffame.dk>
 *
 * @section LICENSE
 * This program is licensed under LGPL3, see LICENSE
 * 
 * @section DESCRIPTION
 * This file has the annotation class
 */
#ifndef ECGLIB_ANNOTATION_LJ_2014_03_27
#define ECGLIB_ANNOTATION_LJ_2014_03_27 1

#include <ecglib_bits/util/util.hpp>
#include <boost/enum.hpp>

#include <armadillo>

#include <ecglib_bits/ecglib.hpp>

#include <algorithm>

#include <vector>
#include <string>

#include <map>

namespace ecglib {
	using namespace arma;

	const int GLOBAL_LEAD=-1;
	typedef int leadnumber;

	/**
	* @brief ECG annotation
	* @todo Instead of saving location as an unsigned int, it should use boost::unit or something, so we know the original unit (or unitless)
	*/
	class annotation {
		// Constructors
		public:

			/**
			* @brief Empty annotation
			* @todo is this constructor really needed ?
			*/
			annotation() {}

			/**
			* @brief Create annotation in global lead
			*
			* @param location Location of annotation
			* @param type Annotation type
			*/
			annotation(unsigned int location, const annotation_type type) : _location(location), _type(type),_lead(GLOBAL_LEAD), _subtype(annotation_subtype::NONE) {}

			/**
			* @brief Create an annotation in a specific lead
			*
			* @param location Location of annotation
			* @param type Annotation type
			* @param l Annotation lead
			*/
			annotation(unsigned int location, const annotation_type type, const leadnumber l) : _location(location), _type(type), _lead(l), _subtype(annotation_subtype::NONE) {}

			/**
			* @brief Create an annotation in a specific lead
			*
			* @param location Location of annotation
			* @param type Annotation type
			* @param l Annotation lead
			* @param sub Subannotation type
			*/
			annotation(unsigned int location, const annotation_type type, const leadnumber l, const annotation_subtype sub) : _location(location), _type(type), _lead(l), _subtype(sub) {}

		// Methods:
		public:
			bool operator<(const annotation &other) const {
				return _location < other._location;
			}

			bool operator>(const annotation &other) const {
				return _location > other._location;
			}

			bool operator<(const unsigned int &other) const {
				return _location < other;
			}

			bool operator>(const unsigned int &other) const {
				return _location > other;
			}

			annotation operator-(const unsigned int subt) const {
				return annotation(_location-subt,_type,_lead);
			}

			operator unsigned int() const {
				return _location;
			}

		// Setters/getters:
		public:

			/**
			* @brief Get ecglead of annotation
			*
			* @return ecglead
			*/
			leadnumber lead() const {
				return _lead;
			}

			/**
			* @brief Set ecglead for annotation
			*
			* @param lead ecglead
			*/
			void lead(leadnumber lead) {
				_lead = lead;
			}

			/**
			* @brief Get location of annotation
			*
			* @return location location of annotation
			*/
			unsigned int location() const {
				return _location;
			}

			/**
			* @brief Set location of annotation
			*
			* @param location location to set
			*/
			void location(unsigned int location) {
				_location = location;
			}

			/**
			* @brief Get type of annotation
			*
			* @return Annotation type
			*/
			annotation_type type() const {
				return _type;
			}

			/**
			* @brief Get annotation type
			*
			* @param type Annotation type
			*/
			void type(annotation_type type) {
				_type = type;
			}

			/**
			* @brief Get annotation subtype (used to represent beat labels eg.)
			*
			* @return annotation subtype
			*/
			annotation_subtype subtype() const {
				return _subtype;
			}

			/**
			* @brief Set annotation subtype
			*
			* @param subtype Annotation subtype
			*/
			void subtype(annotation_subtype subtype) {
				_subtype = subtype;
			}

		// Helpers:
		private:

		// Attributes:
		protected:
			/**
			 * @brief Location in ms
			 */
			unsigned int _location;

			annotation_type _type;
			leadnumber _lead;
			annotation_subtype _subtype;
	};

	/**
	 * @brief Annotationset class - contains annotations for a lead
	 */
	class annotationset {
		public:
			typedef std::map<unsigned int, annotation>::iterator iterator;
			typedef std::map<unsigned int, annotation>::const_iterator const_iterator;

		public:
			annotationset() {
			}

		public:
			annotation& operator[](const unsigned int idx) {
				return _annset[idx];
			}

			const annotation& operator[](const unsigned int idx) const {
				const_iterator iter = _annset.find(idx);

				if(iter == _annset.end()) {
					std::stringstream ss;
					ss << "No annotation at:" << idx;
					throw std::logic_error(ss.str());
				}

				return iter->second;
			}

			iterator begin() {
				return _annset.begin();
			}

			iterator end() {
				return _annset.end();
			}

			iterator lower_bound(const unsigned int idx) {
				return _annset.lower_bound(idx);
			}

			iterator upper_bound(const unsigned int idx) {
				return _annset.upper_bound(idx);
			}

			const_iterator begin() const {
				return _annset.begin();
			}

			const_iterator end() const {
				return _annset.end();
			}

			const_iterator lower_bound(const unsigned int idx) const {
				return _annset.lower_bound(idx);
			}

			const_iterator upper_bound(const unsigned int idx) const {
				return _annset.upper_bound(idx);
			}

			iterator find(unsigned int idx) {
				return _annset.find(idx);
			}

			const_iterator find(unsigned int idx) const {
				return _annset.find(idx);
			}

			std::size_t size() const {
				return _annset.size();
			}

			void clear() {
				_annset.clear();
			}

			std::size_t erase(unsigned int idx) {
				iterator it = _annset.find(idx);

				if(it == _annset.end()) {
					return 0;
				}

				erase(it);

				return 1;
			}

			bool erase(iterator it) {
				_annset.erase(it);

				return 1;
			}

			template<typename PRED>
			iterator remove_if(PRED &pred) {
				iterator it = begin();
				iterator endit = end();

				while(it != endit) {
					if(pred(*it)) {
						erase(it++);
					} else {
						++it;
					}
				}

				return it;
			}

			template<typename INPUT_ITER>
			iterator insert(INPUT_ITER a1, INPUT_ITER a2) {
				iterator ret;

				while(a1 != a2) {
					ret = _annset.insert(*a1).first;

					++a1;
				}

				return ret;
			}

		private:
			std::map<unsigned int, annotation> _annset;
	};

	/**
	 * @brief Pointmap, links annotationsets with ecgleads
	 */
	class pointmap {
		public:
			typedef std::map<leadnumber, annotationset>::iterator iterator;
			typedef std::map<leadnumber, annotationset>::const_iterator const_iterator;

		public:
			pointmap() {
			}

		public:
			annotationset& operator[](const leadnumber l) {
				return _pm[l];
			}

			const annotationset& operator[](const leadnumber l) const {
				const_iterator iter = _pm.find(l);

				if(iter == _pm.end()) {
					std::stringstream ss;
					ss << "No such lead number:" << l;
					throw std::logic_error(ss.str());
				}

				return iter->second;
			}

			iterator begin() {
				return _pm.begin();
			}

			iterator end() {
				return _pm.end();
			}

			const_iterator begin() const {
				return _pm.begin();
			}

			const_iterator end() const {
				return _pm.end();
			}

			std::size_t size() const {
				return _pm.size();
			}

			iterator find(ecglib::leadnumber l) {
				return _pm.find(l);
			}

			const_iterator find(ecglib::leadnumber l) const {
				return _pm.find(l);
			}

			void clear() {
				_pm.clear();
			}

			void erase(iterator it) {
				_pm.erase(it);
			}
	
			template<typename INPUT_ITER>
			iterator insert(INPUT_ITER a1, INPUT_ITER a2) {
				iterator ret;

				while(a1 != a2) {
					ret = _pm.insert(*a1);

					++a1;
				}

				return ret;
			}

		private:
			std::map<leadnumber, annotationset> _pm;
	};

	class annotationtype_in : public std::unary_function<annotation, bool> {
		public:
			annotationtype_in(const std::vector<annotation> &list) : _alist(list) {
			}

		public:
			result_type operator()(argument_type in) {
				for(std::size_t i = 0; i < _alist.size(); ++i) {
					if(_alist[i] == in) return true;
				}

				return false;
			}

		private:
			std::vector<annotation> _alist;
	};

	/**
	* @brief Get annotations by type
	*
	* @param pm Annotation set
	* @param typ Annotation type
	* @param out Vector of annotations
	*/
	void get_annotations(const annotationset &pm, const annotation_type &typ, std::vector<annotation> &out);

	/**
	 * @brief Get annotations across leads by type
	 *
	 * @param pm Pointmap
	 * @param typ Annotation type
	 * @param out Vector of annotations
	 */
	void get_annotations(const pointmap &pm, const annotation_type &typ, std::vector<annotation> &out);

	/**
	 * @brief Shift all annotations to a new start, used for chopping an ECG
	 *
	 * @param ann Annotationset
	 * @param newstart New start
	 *
	 * @return New annotationset
	 */
	annotationset rebase(const annotationset &ann, const int newstart);

	/**
	 * @brief Shifts all annotations to a new start, used for chopping an ECG
	 *
	 * @param pm Pointmap
	 * @param newstart New start
	 */
	void rebase(pointmap &pm, const int newstart);
}

#endif
