/**
 * @file ecgdata.hpp
 * @author Lars Johannesen <lj@halloffame.dk>
 *
 * @section LICENSE
 * This program is licensed under LGPL3, see LICENSE
 * 
 * @section DESCRIPTION
 * The classes in this file are used for describing characteristics about ecg records
 */
#ifndef ECGLIB_ECGDATA_LJ_2014_03_27
#define ECGLIB_ECGDATA_LJ_2014_03_27 1

#include <ecglib_bits/util/util.hpp>

#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/enum.hpp>
#include <boost/assert.hpp>

#include <boost/bimap.hpp>

#include <armadillo>

#include <ecglib_bits/annotation.hpp>
#include <ecglib_bits/ecglib.hpp>

#include <algorithm>

#include <vector>
#include <string>

#include <map>

namespace ecglib {
	using namespace arma;

	/**
	* @brief ECGdata class, contains a matrix (nsamples x nleads), vector of lead names and annotations by lead
	* @todo Change internals so annotations are not in samples and resolution does not have to be uV.. 
	*/
	template<class T>
	class Ecgdata {
		public:
		// NOTE: Loops all samples for a leads - lj
		typedef mat::col_iterator leaditerator;
		typedef mat::const_col_iterator const_leaditerator;

		// NOTE: Loops all samples for all leads - lj
		typedef mat::row_iterator sampleiterator;
		typedef mat::const_row_iterator const_sampleiterator;

		typedef ecglib::annotationset annotationset;
		typedef ecglib::pointmap pointmap;
		typedef pointmap::iterator pointmapiterator;
		typedef pointmap::const_iterator const_pointmapiterator;

		typedef annotationset::iterator pointiterator;
		typedef annotationset::const_iterator const_pointiterator;

		typedef boost::bimap<leadnumber, ecglead> leadmap;

		// Constructors
		public:

			/**
			* @brief Creates empty ecgdata
			* @todo This is needed because of the way we load signals currently. Should be changed imho, as an empty ecgdata is kind of a "zombie" object / invalid state
			*/
			Ecgdata() : _nsamples(0), _nleads(0) {}
		
			/**
			* @brief Creates an ecgdata object of length nsamples, but number of leads based on ecgheader information
			*
			* @param nsamples Number of samples (could be equal of bigger than nsample  from ecgheader
			* @param ecgheader 
			*/
            		Ecgdata(unsigned int nsamples, ecgheader eh) : _data(nsamples,eh.nleads), _nsamples(nsamples) {
			//TODO: populate leads casting from the ecgheader.leadnames
			}

			/**
			* @brief Creates an ecgdata object of length nsamples, number of columns/leads corresponding to leadnames size
			*
			* @param nsamples Number of samples
			* @param leadnames Lead names of the leads
			*/
			Ecgdata(unsigned int nsamples, const std::vector<ecglead> &leadnames) : _data(nsamples,leadnames.size()), _nsamples(nsamples), _nleads(leadnames.size()) { 
				typedef leadmap::value_type lval;

				for(std::size_t i = 0; i < leadnames.size(); ++i) {
					_leadmap.insert(lval(i, leadnames[i]));
				}
			}

			Ecgdata(unsigned int nsamples, unsigned int nleads) : _data(nsamples, nleads), _nsamples(nsamples), _nleads(nleads) {
			}

			Ecgdata(const mat &indata, const leadmap &lm) : _data(indata), _leadmap(lm), _nsamples(indata.n_rows), _nleads(indata.n_cols) {
			}

			Ecgdata(const mat &indata, const double fs) : _data(indata), _fs(fs), _nsamples(indata.n_rows), _nleads(indata.n_cols) {
			}

			/**
			* @brief Creates an ecgdata class from a matrix of data and a set of lead names (first column is first element in leadnames
			*
			* @param indata Matrix of data
			* @param leadnames Lead names of the data
			*/
			Ecgdata(const mat &indata, const std::vector<ecglead> &leadnames) : _data(indata), _nsamples(indata.n_rows), _nleads(leadnames.size()) {
				if(indata.n_cols != _nleads) {
					throw ecglib::ecglib_exception("ecglib::constructor::Length of leadnames does not match column count");
				}

				typedef leadmap::value_type lval;

				for(std::size_t i = 0; i < leadnames.size(); ++i) {
					_leadmap.insert(lval(i, leadnames[i]));
				}
			}

		// Public methods
		public:

			/**
			* @brief Gets the column number of the lead. Throws an exception if the lead is not in the data
			*
			* @param lead Lead to get the column number for
			*
			* @return Column number
			*/
			int leadnum(const ecglead lead) const {
				leadmap::right_const_iterator rit = _leadmap.right.find(lead);

				if(rit == _leadmap.right.end()) {
					throw ecglib::ecglib_exception("No such lead");
				}

				return rit->second;
			}

			/**
			* @brief Get leadname for a column number
			*
			* @param leadnum Column number
			*
			* @return lead name
			*/
			ecglead leadname(int leadnum) const {
				leadmap::left_const_iterator lit = _leadmap.left.find(leadnum);

				if(lit == _leadmap.left.end()) {
					throw ecglib::ecglib_exception("Lead has no lead name");
				}

				return lit->second;
			}

			/**
			* @brief Adds a lead to the container. Causes container to resize internal storage (not efficient)
			*
			* @tparam ITER An iterator, must be usable by std::copy
			* @param lead Lead name for the lead to add
			* @param start Start of what to copy
			* @param stop Stop of what to copy
			*/
			template<class ITER>
			void add_lead(const ecglead lead, ITER start, ITER stop) {
				std::size_t length = stop-start;
				int newlead = _nleads;

				if(_nsamples == 0) {
					_nsamples = length;
				} else if(_nsamples != length) {
					throw ecglib::ecglib_exception("ecglib::add_lead: _nsamples!=length");
				} else {
					leadmap::right_iterator rit = _leadmap.right.find(lead);
					
					if(rit != _leadmap.right.end()) {
						throw ecglib::ecglib_exception("ecglib::add_lead:lead is not new");
					}
				}
			
				_data.reshape(_nsamples,_nleads+1);

				std::copy(start,stop,_data.begin_col(_nleads));

				leadmap::value_type lval(newlead, lead);
				_leadmap.insert(lval);
				++_nleads;
			}

		// Setters / Getters
		public:
			/**
			* @brief Gets subject id
			*
			* @return subject id
			*/
			std::string id() const {
				return _id;
			}

			/**
			* @brief Sets subject id
			*
			* @param id2 Subject id
			*/
			void id(const std::string &id2) {
				_id = id2;
			}

			/**
			* @brief Get gender of subject (enum)
			*
			* @return Gender
			*/
			gender sex() const {
				return _sex;
			}

			/**
			* @brief Set gender of subject
			*
			* @param sex2 Gender
			*/
			void sex(const gender sex2) {
				_sex = sex2;
			}

			/**
			* @brief Get birthdate
			*
			* @return Birthdate
			*/
			boost::gregorian::date birthdate() const {
				return _birthdate;
			}

			/**
			* @brief Set birthdate
			*
			* @param bday Birthdate
			*/
			void birthdate(const boost::gregorian::date &bday) {
				_birthdate = bday;
			} 

			/**
			* @brief Get start of recording
			*
			* @return Start of recording
			*/
			boost::posix_time::ptime recordstart() const {
				return _recordstart;
			}

			/**
			* @brief Set start of recording
			*
			* @param pt Start of recording
			*/
			void recordstart(const boost::posix_time::ptime &pt) {
				_recordstart = pt;
			}

			/**
			* @brief Get sampling frequency
			*
			* @return Sampling frequency
			* @todo Should be stored in proper units
			*/
			double fs() const {
				return _fs;
			}

			/**
			* @brief Set sampling frequency
			*
			* @param fs2 Sampling frequency
			*/
			void fs(const double fs2) {
				_fs = fs2;
			}

			/**
			* @brief Get resolution
			*
			* @return Resolution how many units per uV, i.e. if data is in uV it is 1
			*/
			double resolution() const {
				return _res;
			}

			/**
			* @brief Set resolution
			*
			* @param res2 resolution
			*/
			void resolution(const double res2) {
				_res = res2;
			}

			/**
			* @brief Get number of samples
			*
			* @return Nsamples
			*/
			std::size_t nsamples() const {
				return _nsamples;
			}
	
			/**
			* @brief Get number of leads
			*
			* @return Nleads
			*/
			std::size_t nleads() const {
				return _nleads;
			}

			/**
			* @brief Get annotations for data
			*
			* @return annotations
			*/
			pointmap pointsmap() const {
				return _points;
			}

			/**
			* @brief Set annotaitons for data
			*
			* @param pts annotations
			*/
			void pointsmap(const pointmap &pts) {	
				_points = pts;
			}

			/**
			* @brief Get vector of leads stored
			*
			* @return Vector leads
			*/
			leadmap leadnames() const {
				return _leadmap;
			}

			/**
			* @brief Get vector of leads stored
			*
			* @return Vector leads
			*/
			void leadnames(const leadmap &lm) {
				_leadmap = lm;
			}

			/**
			* @brief Insert annotation
			*
			* @param a annotation
			*/
			void insert_annotation(const annotation &a) {
				_points[a.lead()][a.location()] = a;
			}

			/**
			* @brief Remove annotation
			*
			* @param a annotation
			*/
			void remove_annotation(const annotation &a) {
				pointmapiterator pm = _points.find(a.lead());

				if(pm == _points.end()) {
					throw ecglib::ecglib_exception("No points for that lead");
				}

				pointiterator pi = _points[a.lead()].find(a.location());

				if(pi != _points[a.lead()].end()) {
					_points[a.lead()].erase(pi);
				}
			}

		// Container methods
		public:
			/**
			* @brief Get sample reference by lead,sample
			*
			* @param lead lead
			* @param sample Sample number
			*
			* @return sample reference
			*/
			double& operator()(const ecglead &lead, const int sample) {
				int leadn = leadnum(lead);		

				return _data(sample,leadnum(leadn));
			}

			/**
			* @brief Get sample value by lead,sample
			*
			* @param lead lead
			* @param sample Sample number
			*
			* @return sample value
			*/
			double operator()(const ecglead &lead, const int sample) const {
				int leadn = leadnum(lead);		

				return _data(sample,leadnum(lead));
			}

			const subview_col<double> operator()(const span& rowspan, const uword colnum) const {
				return _data(rowspan, colnum);
			}

			/**
			* @brief Get sample reference by lead,sample
			*
			* @param lead lead number
			* @param sample sample
			*
			* @return sample reference
			*/
			double& operator()(const int lead, const int sample) {
				return _data(sample,lead);
			}

			/**
			* @brief Get sample value by lead,sample
			*
			* @param lead lead number
			* @param sample sample
			*
			* @return sample value
			*/
			double operator()(const int lead, const int sample) const {
				return _data(sample,lead);
			}

			/**
			* @brief Get lead by number
			*
			* @param lead lead number
			*
			* @return lead (subview)
			*/
			subview_col<double> lead(const ecglead lead) {
				int leadn = leadnum(lead);		

				return _data.col(leadn);
			}

			/**
			* @brief Const-correct get lead by number
			*
			* @param lead lead number
			*
			* @return lead (subview) constant
			*/
			const subview_col<double> lead(const ecglead lead) const {
				int leadn = leadnum(lead);		

				return _data.col(leadn);
			}

			/**
			* @brief Get lead by number
			*
			* @param lead lead number
			*
			* @return lead (subview)
			*/
			subview_col<double> lead(const int lead) {
				return _data.col(lead);
			}

			/**
			* @brief Const-correct get lead by number
			*
			* @param lead lead number
			*
			* @return lead (subview) constant
			*/
			const subview_col<double> lead(const int lead) const {
				return _data.col(lead);
			}

			/**
			* @brief Get part of a lead
			*
			* @param lead lead number
			* @param start start sample
			* @param stop stop sample
			*
			* @return part of lead 
			*/
			subview_col<double> lead(const int lead, const int start, const int stop) {
				return _data(span(start,stop),lead);
			}

			/**
			* @brief Get part of a lead, const-correct
			*
			* @param lead lead number
			* @param start start sample
			* @param stop stop sample
			*
			* @return part of lead constant
			*/
			const subview_col<double> lead(const int lead, const int start, const int stop) const {
				return _data(span(start,stop),lead);
			}

			/**
			* @brief Get part of a lead
			*
			* @param lead lead number
			* @param start start sample
			* @param stop stop sample
			*
			* @return part of lead 
			*/
			subview_col<double> lead(const ecglead lead, const int start, const int stop) {
				int leadn = leadnum(lead);		

				return _data(span(start,stop),leadn);
			}

			/**
			* @brief Get part of a lead, const-correct
			*
			* @param lead lead number
			* @param start start sample
			* @param stop stop sample
			*
			* @return part of lead constant
			*/
			const subview_col<double> lead(const ecglead lead, const int start, const int stop) const {
				int leadn = leadnum(lead);

				return _data(span(start,stop),leadn);
			}

			/**
			* @brief get data or matrix of all data
			*
			* @return data
			* @todo remove?
			*/
			subview<double> data() {
				return _data(span::all,span::all);
			}

			/**
			* @brief get data or matrix of all data, const-correct
			*
			* @return data const
			* @todo remove?
			*/
			subview<double> data() const {
				return _data(span::all,span::all);
			}

			/**
			* @brief Begin lead iterator
			*
			* @param lead lead
			*
			* @return lead iterator
			*/
			leaditerator begin_lead(const ecglead &lead) {
				return _data.begin_col(leadnum(lead));
			}

			/**
			* @brief End lead iterator
			*
			* @param lead lead
			*
			* @return lead iterator
			*/
			leaditerator end_lead(const ecglead &lead) {
				return _data.end_col(leadnum(lead));
			}

			/**
			* @brief Begin lead iterator
			*
			* @param lead lead
			*
			* @return lead iterator const
			*/
			const_leaditerator begin_lead(const ecglead &lead) const {
				return _data.begin_col(leadnum(lead));
			}

			/**
			* @brief End lead iterator
			*
			* @param lead lead
			*
			* @return lead iterator const
			*/
			const_leaditerator end_lead(const ecglead &lead) const {
				return _data.end_col(leadnum(lead));
			}

			/**
			* @brief Begin lead iterator
			*
			* @param lead lead number
			*
			* @return lead iterator
			*/
			leaditerator begin_lead(int num) {
				return _data.begin_col(num);
			}

			/**
			* @brief Begin lead iterator
			*
			* @param lead lead number
			*
			* @return lead iterator
			*/
			leaditerator end_lead(int num) {
				return _data.end_col(num);
			}

			/**
			* @brief Begin lead iterator
			*
			* @param lead lead number
			*
			* @return lead iterator const
			*/
			const_leaditerator begin_lead(int num) const {
				return _data.begin_col(num);
			}

			/**
			* @brief End lead iterator
			*
			* @param lead lead number
			*
			* @return lead iterator const
			*/
			const_leaditerator end_lead(int num) const {
				return _data.end_col(num);
			}

			/**
			* @brief Begin sample (across leads)
			*
			* @param sample Sample number
			*
			* @return sample iterator
			*/
			sampleiterator begin_sample(int sample) {
				return _data.begin_row(sample);
			}

			/**
			* @brief End sample (across leads)
			*
			* @param sample Sample number
			*
			* @return sample iterator
			*/
			sampleiterator end_sample(int sample) {
				return _data.end_row(sample);
			}

			/**
			* @brief Begin sample (across leads)
			*
			* @param sample Sample number
			*
			* @return sample iterator const
			*/
			const_sampleiterator begin_sample(int sample) const {
				return _data.begin_row(sample);
			}

			/**
			* @brief End sample (across leads)
			*
			* @param sample Sample number
			*
			* @return sample iterator const
			*/
			const_sampleiterator end_sample(int sample) const {
				return _data.end_row(sample);
			}

			/**
			* @brief Get begin pointmap const iterator
			*
			* @param lead lead
			*
			* @return iterator to end
			*/
			pointmapiterator pointmap_begin() {
				return _points.begin();
			}

			/**
			* @brief Get end pointmap iterator
			*
			* @param lead lead
			*
			* @return iterator to end
			*/
			pointmapiterator pointmap_end() {
				return _points.end();
			}

			/**
			* @brief Get begin pointmap const iterator
			*
			* @param lead lead
			*
			* @return iterator to begin
			*/
			const_pointmapiterator pointmap_begin() const {
				return _points.begin();
			}

			/**
			* @brief Get end pointmap const iterator
			*
			* @param lead lead
			*
			* @return iterator to end
			*/
			const_pointmapiterator pointmap_end() const {
				return _points.end();
			}

			/**
			* @brief Get begin-point iterator by lead
			*
			* @param lead lead
			*
			* @return iterator to begin
			*/
			pointiterator points_begin(const ecglead &lead) {
				int leadn = leadnum(lead);

				pointmapiterator iter = _points.find(leadn);

				if(iter == _points.end()) throw std::runtime_error("Could not find annotations for lead");
	
				return iter->second.begin();
			}

			/**
			* @brief Get begin-point iterator by lead
			*
			* @param lead lead
			*
			* @return iterator to begin
			*/
			pointiterator points_begin(const leadnumber leadn) {
				pointmapiterator iter = _points.find(leadn);

				if(iter == _points.end()) throw std::runtime_error("Could not find annotations for lead");
	
				return iter->second.begin();
			}

			/**
			* @brief Get end-point iterator by lead
			*
			* @param lead lead
			*
			* @return iterator to begin
			*/
			pointiterator points_end(const ecglead &lead) {
				int leadn = leadnum(lead);

				pointmapiterator iter = _points.find(leadn);

				if(iter == _points.end()) throw std::runtime_error("Could not find annotations for lead");

				return iter->second.end();
			}

			/**
			* @brief Get end-point iterator by lead
			*
			* @param lead lead
			*
			* @return iterator to begin
			*/
			pointiterator points_end(const leadnumber leadn) {
				pointmapiterator iter = _points.find(leadn);

				if(iter == _points.end()) throw std::runtime_error("Could not find annotations for lead");

				return iter->second.end();
			}

			/**
			* @brief Get begin-point iterator by lead
			*
			* @param lead lead
			*
			* @return const iterator to begin
			*/
			const_pointiterator points_begin(const leadnumber leadn) const {
				const_pointmapiterator iter = _points.find(leadn);

				if(iter == _points.end()) throw std::runtime_error("Could not find annotations for lead");
	
				return iter->second.begin();
			}

			/**
			* @brief Get end-point iterator by lead
			*
			* @param lead lead
			*
			* @return const iterator to end
			*/
			const_pointiterator points_end(const leadnumber leadn) const {
				const_pointmapiterator iter = _points.find(leadn);

				if(iter == _points.end()) throw std::runtime_error("Could not find annotations for lead");

				return iter->second.end();
			}

			/**
			* @brief Get begin-point iterator by lead
			*
			* @param lead lead
			*
			* @return const iterator to begin
			*/
			const_pointiterator points_begin(const ecglead &lead) const {
				int leadn = leadnum(lead);

				const_pointmapiterator iter = _points.find(leadn);

				if(iter == _points.end()) throw std::runtime_error("Could not find annotations for lead");
	
				return iter->second.begin();
			}

			/**
			* @brief Get end-point iterator by lead
			*
			* @param lead lead
			*
			* @return const iterator to end
			*/
			const_pointiterator points_end(const ecglead &lead) const {
				int leadn = leadnum(lead);

				const_pointmapiterator iter = _points.find(leadn);

				if(iter == _points.end()) throw std::runtime_error("Could not find annotations for lead");

				return iter->second.end();
			}

			/**
			* @brief Extracts a sub-part of an ecgdata class
			*
			* @param start First sample
			* @param stop Last sample
			*
			* @return ECGData of the sub part (both data/annotations)
			*/
			Ecgdata<T> subpart(int starttime, int stoptime) const {
				if (stoptime <= 0) {
					unsigned int length = static_cast<unsigned int>(round( (_nsamples/_fs)*1000. ));

					stoptime += length-1;
				}
				unsigned int start = static_cast<unsigned int>(round((starttime/1000.) * _fs)); 
				unsigned int stop = static_cast<unsigned int>(round((stoptime/1000.) * _fs)); 

				if(start < 0 || stop > _nsamples || start > stop) {
					throw ecglib::ecglib_exception("Start_or_stop_incorrect");
				}

				Ecgdata<T> e(_data(span(start,stop),span::all), _leadmap);

				e.fs( _fs);
				e.resolution(_res);
				e.id(_id);
				e.birthdate(_birthdate);
				e.recordstart(_recordstart); //FIXME: Should update this accordingly - lj

				// Copy annotations
				pointmap pm;

				const_pointmapiterator pmend = _points.end();
				for(const_pointmapiterator pmi = _points.begin(); pmi != pmend; ++pmi) {
					const_pointiterator pistart = pmi->second.lower_bound(starttime);
					const_pointiterator pistop = pmi->second.upper_bound(stoptime);

					annotationset as;

					while(pistart != pistop) {
						as[pistart->first-starttime] = pistart->second-static_cast<unsigned int>(starttime);

						++pistart;
					}

					pm[pmi->first] = as;
				}

				e.pointsmap(pm);

				return e; 
			}

			/**
			 * @brief Extracts a sub-part of an ecgdata class
			 *
			 * @param leads ECG leads
			 *
			 * @return ECGdata
			 */
			Ecgdata<T> subpart(const std::vector<ecglib::ecglead> &leads) const {
				mat ecg = zeros<mat>(_nsamples+1, leads.size());

				for(std::size_t i = 0; i < leads.size(); ++i) {
					std::copy(begin_lead(leads[i]),end_lead(leads[i]), ecg.begin_col(i));
				}

				Ecgdata<T> e(ecg, leads);
				e.fs( _fs);
				e.resolution(_res);
				e.id(_id);
				e.birthdate(_birthdate);
				e.recordstart(_recordstart); //FIXME: Should update this accordingly - lj
				Ecgdata<T>::pointmap pm;
				typedef Ecgdata<T>::const_pointmapiterator pmiter;
				pmiter pmend = _points.end();

				typedef std::vector<ecglead>::const_iterator liter;
				liter lend = leads.end();

				for(pmiter pmi = _points.begin(); pmi != pmend; ++pmi) {
					if(pmi->first == ecglib::GLOBAL_LEAD) {
						pm[pmi->first] = pmi->second;
					} else {
						for(liter li = leads.begin(); li != lend; ++li) {
							if(*li == pmi->first) {
								pm[pmi->first] = pmi->second;
							}
						}
					}
				}

				e.pointsmap(pm);

				return e;
			}

		// Helpers
		private:

		// Attributes
		protected:
			Mat<T> _data;
			std::size_t _nsamples;
			leadmap _leadmap;
			double _fs;
			double _res;
			std::size_t _nleads;

			std::string _id;
			boost::gregorian::date _birthdate;
			gender _sex;
			boost::posix_time::ptime _recordstart;

			pointmap _points;
	};

	typedef Ecgdata<double> ecgdata;
}

#endif
