/**
 * @file detector.hpp
 * @author Lars Johannesen <lj@halloffame.dk>
 *
 * @section LICENSE
 * This program is licensed under LGPL3, see LICENSE
 * 
 * @section DESCRIPTION
 * The classes in this file are used for describing characteristics about ecg records
 */
#ifndef ECGLIB_DETECTOR_LJ_2014_06_06
#define ECGLIB_DETECTOR_LJ_2014_06_06 1

#include <ecglib_bits/annotation.hpp>

#include <armadillo>

#include <tuple>

namespace ecglib {
	/**
	 * @brief Find center of gravity, left / right of peak and area under the curve for the peak
	 *
	 * @tparam T Signal t
	 * @param sig Signal vector
	 * @param frac Fraction to use for detecting left/right 
	 * @param sfrac Fraction to use for detecting left/right (slope)
	 * @param cpeak Candidate peak
	 * @param tuple <location, left, right>
	 */
	template<typename T>
	std::tuple<
		unsigned int,
		unsigned int,
		unsigned int
	>
	findcog(const T& sig, double frac, double sfrac, unsigned int cpeak) {
		unsigned int loc;
		unsigned int left, right;
		left = right = 0;
	
		unsigned int valleft,valright;
		valleft = valright = cpeak;

		double leftsum, rightsum;
		leftsum = rightsum = 0;

		// Find left cutoff
		double maxleftslop = 0;
		int maxleftsloploc = 0;
		double thr = sig(cpeak) * frac;
		double minleftval = 0;
		bool found = false;
		while(valleft > 1) {
			if(!found) minleftval = sig(valleft);
			else if (minleftval > sig(valleft)) minleftval = sig(valleft);

			double slop = sig(valleft+1)-sig(valleft);
			if (slop > maxleftslop) {
				maxleftslop = slop;
				maxleftsloploc = valleft;
			}

			if(sig(valleft) < thr) break;

			leftsum += sig(valleft);
			--valleft;
		}

		// Find right cutoff
		double minrightval = 0;
		double maxrightslop = 0;
		int maxrightsloploc = 0;
		found = false;
		while(valright < (sig.n_elem)-1) {
			if(!found) minrightval = sig(valright);
			else if (minrightval > sig(valright)) minrightval = sig(valright);

			double slop = sig(valright)-sig(valright+1);
			if (slop > maxrightslop) {
				maxrightslop = slop;
				maxrightsloploc = valright;
			}

			if(sig(valright) < thr) break;

			rightsum += sig(valright);
			++valright;
		}

		// Define total area
		double thrsum = (leftsum + rightsum + sig(cpeak)) / 2.;
		double tmp = 0;
		leftsum = 0;
		rightsum = 0;
		loc = 0;

		// Find center of gravity
		int i = valleft;
		while(i < valright) {
			tmp += sig(i);

			if (tmp > thrsum) {
				if (loc == 0 ) {
					loc = i;
				} else {
					rightsum += sig(i);
				}
			} else {
				leftsum += sig(i);
			}

			++i;
		}

		int slopleft = valleft;
		while (slopleft > 0) {
			double slop = sig(slopleft+1)-sig(slopleft);

			if (slop < sfrac * maxleftslop) {
				break;
			}

			--slopleft;
		}

		int slopright = valright;
		while (slopright < (sig.n_elem-1)) {
			double slop = sig(slopright)-sig(slopright+1);

			if (slop < sfrac * maxrightslop) {
				break;
			}

			++slopright;
		}

		left = valleft > slopleft ? slopleft : valleft;
		right = valright < slopright ? slopright : valright;

		return std::make_tuple(loc,left,right);
	}

	/**
	* @brief Detector state base
	*/
	class detectorstate {
		public:
			/**
			* @brief Initialize detector state
			*
			* @param nchannels Number of channels
			*/
			detectorstate(const std::size_t nchannels) : _nch(nchannels) {
				_ready.resize(_nch);
				_cps.resize(_nch);

				for(std::size_t i = 0; i < _nch; ++i) {
					_ready[i] = false;
					_cps[i] = 0;
				}	
			}
	
		public:
			/**
			* @brief Set channel ch to be ready
			*
			* @param ch Channel number
			*/
			void ready(const std::size_t ch) {
				_ready[ch] = true;
			}

			/**
			* @brief Get candidate peak for channel ch
			*
			* @param ch Channel number
			*
			* @return Candidate peak
			*/
			std::size_t cpeak(const std::size_t ch) const {
				return _cps[ch];
			}

			/**
			* @brief Set candidate peak for channel ch
			*
			* @param ch Channel ch
			* @param cp Candidate peak
			*/
			void cpeak(const std::size_t ch, const std::size_t cp) {
				_cps[ch] = cp;
				_ready[ch] = false;
			}

			/**
			* @brief Return true if channel ch is ready
			*
			* @param ch Channel ch
			*
			* @return True if ready
			*/
			bool isready(const std::size_t ch) const {
				return _ready[ch];
			}

			/**
			* @brief Return number of ready channels
			*
			* @return Number of ready channels
			*/
			std::size_t nready() const {
				std::size_t n = 0;
				for(bool r : _ready) {
					n += r ? 1 : 0;
				}

				return n;
			}

		protected:
			std::size_t _nch;

			std::vector<bool> _ready;
			std::vector<std::size_t> _cps;
	};

	/**
	* @brief Default detector policy
	*/
	class defaultdetectorpolicy {
		public:
			typedef ecglib::pointmap annotation;

		public:
			/**
			* @brief Initialize default detector
			*
			* @param nchannels Number of channels
			*/
			defaultdetectorpolicy(const std::size_t nchannels)  : _nch(nchannels) {
				_peakloc.resize(_nch);
				_peakready.resize(_nch);

				for(std::size_t i = 0; i < _nch; ++i) {
					_peakready[i] = false;
				}
			}

		public:
			/**
			* @brief Trigger candidate peak detection
			*
			* @tparam data Data class
			* @param state Detector state class
			* @param input Input data
			* @param ch Channel ch
			* @param idx Index of candidate peak
			* @param fs Sampling frequency
			*
			* @return True if a candidate peak is detected
			*/
			template<class data>
			bool cpeak(detectorstate &state, const data &input, const std::size_t ch, const std::size_t idx, const double fs) {
				bool res = false;

				try {
					auto cogres = findcog(input.col(ch), 0.5, state.cpeak(ch));
					std::size_t loc = std::get<0>(cogres);

					if (!_peakready[ch]) {
						_peakready[ch] = true;
						_peakloc[ch] = loc;
						res = true;
					}
				} catch (...) {
					// Note: Nothing needs to happen if it is not possible to determine COG for the basic policy
				}

				return res;
			}

			/**
			* @brief Detect a peak
			*
			* @tparam data Data type
			* @param pm Pointmap
			* @param idx Index of peak
			* @param state Detector state
			* @param input Input data
			* @param fs Sampling frequency
			*
			* @return True if peak is detected
			*/
			template<class data>
			bool peak(annotation &pm, const std::size_t idx, detectorstate &state, const data &input, const double fs) {
				bool res = false;

				for(std::size_t i = 0; i < _nch; ++i) {
					if(_peakready[i]) {
						pm[i][_peakloc[i]] = ecglib::annotation(_peakloc[i],ecglib::annotation_type::RPEAK,i);
						_peakready[i] = false;
						res = true;
					}
				}

				return res;
			}

		protected:
			std::size_t _nch;
			std::vector<std::size_t> _peakloc;
			std::vector<bool> _peakready;
	};

	/**
	* @brief Detect peaks
	*
	* @tparam Data class
	* @tparam Detectorpolicy Detector policy class
	* @param input Input data
	* @param nsamp Number of samples to consider
	* @param nch Number of channels
	* @param fs Sampling frequency
	* @param dstate Detector state
	* @param dpolicy Detector policy
	* @param frac Fraction required to trigger detection
	* @param minrr Refactory period (ms)
	*/
	template<class data, class detectorpolicy>
	std::tuple<
		typename detectorpolicy::annotation, 
		detectorstate, 
		detectorpolicy>
	detect_peaks(
		const data &input,
		const std::size_t nsamp,
		const std::size_t nch,
		const double fs, 
		const detectorstate &dstate, 
		const detectorpolicy &dpolicy,
		const double frac = 0.5,
		const std::size_t minrr = 250
	) {
		// Copy state/policy
		detectorstate state_out = dstate;
		detectorpolicy policy_out = dpolicy;

		// Consts to be moved to a propmap
		const std::size_t minrr_samp = static_cast<std::size_t>(round(minrr * fs/1000.)); 

		// Annotations
		typename detectorpolicy::annotation pm;

		std::size_t i = 0;
		while(i < nsamp) {
			for(std::size_t j = 0; j < nch; ++j) {
				// Case1: Still going up
				if(input(i,j) > input(state_out.cpeak(j),j)) {
					state_out.cpeak(j,i);
				// Case2: Went down enough?
				} else if(input(i,j) < frac * input(state_out.cpeak(j),j) && !state_out.isready(j)) {
					state_out.ready(j);
				// Case3: No other peaks in the window ?
				} else if(state_out.isready(j) && (i - state_out.cpeak(j)) > minrr_samp) {
					// Request policy to decide if this candidate peak
					policy_out.cpeak(state_out, input, j, i, fs);
					state_out.cpeak(j,i);
				}
			}

			policy_out.peak(pm, i, state_out, input, fs);
	
			++i;
		}

		return std::make_tuple(pm, state_out, policy_out);
	}
}

#endif
