/********************************************************************************
 *                                  DISCLAIMER                                  *
 * This software and documentation (the "Software") were developed at the Food  *
 * and Drug Administration (FDA) by employees of the Federal Government in the  *
 * course of their official duties. Pursuant to Title 17, Section 105 of the    *
 * United States Code, this work is not subject to copyright protection and is  *
 * in the public domain. Permission is hereby granted, free of charge, to any   *
 * person obtaining a copy of the Software, to deal in the Software without     *
 * restriction, including without limitation the rights to use, copy, modify,   *
 * merge, publish, distribute, sublicense, or sell copies of the Software or    *
 * derivatives, and to permit persons to whom the Software is furnished to do   *
 * so. FDA assumes no responsibility whatsoever for use by other parties of     *
 * the Software, its source code, documentation or compiled executables, and    *
 * makes no guarantees, expressed or implied, about its quality, reliability,   *
 *  or any other characteristic. Further, use of this code in no way implies    *
 * endorsement by the FDA or confers any advantage in regulatory decisions.     *
 * Although this software can be redistributed and/or modified freely, we       *
 * ask that any derivative works bear some notice that they are derived from    *
 * it, and any modified versions bear some notice that they have been modified. * 
 *                                                                              *
 * Written by:                                                                  *
 *  Lars Johannesen                                                             *
 ********************************************************************************/
#include "mex.h"

#include <wfdb/wfdb.h>

#include <vector>
#include <map>
#include <string>

#include <iostream>

#include <stdexcept>

#include <armadillo>

template<typename TYP>
TYP get_matlab_scalar(const mxArray *a) {
	TYP *t = (TYP*)mxGetData(a);
	return *t;
}

void mexFunction (int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
	if(nrhs != 2) {
		mexErrMsgTxt("Need one arguments: record filename");
	}

	try {
		if(!mxIsCell(prhs[0]))
                        mexErrMsgTxt("First argument must be a cell");

		if(!mxIsChar(prhs[1]))
                        mexErrMsgTxt("Second argument must be a char");

		char *rec = mxArrayToString(prhs[1]);
		std::string onam = std::string(rec) + std::string(".dat");

		mwSize *dims = (mwSize*)mxGetDimensions(prhs[0]);
		int nsig = dims[0];

		WFDB_Siginfo *s;

		s = (WFDB_Siginfo*) malloc(nsig * sizeof(WFDB_Siginfo));

		int nsamp = 0;

		using namespace arma;

		int fs = 0;

		std::vector<vec> data;
		for(int i = 0; i < nsig; ++i) {
			mxArray *cell = mxGetCell(prhs[0], i);

			if(!mxIsStruct(cell)) {
				mexErrMsgTxt("Not struct");
			}

			int fieldnum = mxGetFieldNumber(cell, "name");
			mxArray *field = mxGetFieldByNumber(cell, 0, fieldnum);
			s[i].desc = strdup(mxArrayToString(field));

			fieldnum = mxGetFieldNumber(cell, "baseline");
			field = mxGetFieldByNumber(cell, 0, fieldnum);
			s[i].baseline = get_matlab_scalar<double>(field);

			fieldnum = mxGetFieldNumber(cell, "gain");
			field = mxGetFieldByNumber(cell, 0, fieldnum);
			s[i].gain = get_matlab_scalar<double>(field);

			fieldnum = mxGetFieldNumber(cell, "units");
			field = mxGetFieldByNumber(cell, 0, fieldnum);
			if (field) {
				s[i].units = strdup(mxArrayToString(field));
			} else {
				s[i].units = strdup("");
			}

			fieldnum = mxGetFieldNumber(cell, "fs");
			field = mxGetFieldByNumber(cell, 0, fieldnum);
			int tmpfs = get_matlab_scalar<double>(field);

			s[i].bsize = 0;
			s[i].spf = 0;
			s[i].adcres = 0;
			s[i].adczero = 0;

			s[i].cksum = 0;

			s[i].fname = strdup(onam.c_str());
			s[i].fmt = 16;
			s[i].group = 0;

			fieldnum = mxGetFieldNumber(cell, "rawdata");
			field = mxGetFieldByNumber(cell, 0, fieldnum);
			double *ptr = (double*)mxGetPr(field);
			int rows = mxGetM(field);
			int cols = mxGetN(field);

			if(rows != 1 && cols != 1) {
				mexErrMsgTxt("Must be vector");
			}
			int nelem = rows * cols;

			vec d(ptr, nelem, false);

			data.push_back(d);
			s[i].nsamp = nelem;

			if(nsamp == 0) {
				nsamp = nelem;
				fs = tmpfs;
			} else {
				if (nsamp != nelem) {
					mexErrMsgTxt("Nsamp: Must all be same");
				}
				if (fs != tmpfs) {
					mexErrMsgTxt("Fs: Must all be same");
				}
			}
		}

		if(osigfopen(s, nsig) < nsig) {
			mexErrMsgTxt("Cannot write");
		}

		setsampfreq(fs);

		WFDB_Sample *v = (WFDB_Sample*)malloc(sizeof(WFDB_Sample)*nsig);

		for(int i = 0; i < nsamp; ++i) {
			for(int j = 0; j < nsig; ++j) {
				v[j] = data[j][i];
			}

			if(i == 0) {
				for(int j = 0; j < nsig; ++j) {
					s[j].initval = v[j];
				}
			}

			if(putvec(v) < 0) {
				mexErrMsgTxt("cannot write all");
			}
		}

		int res = newheader(rec);
		if (res!=0) {
			mexErrMsgTxt("Cant write header");
		}

		wfdbquit();
	} catch (...) {
		mexErrMsgTxt("Failed loading");
	}
}
